/* 
 *
 * $Id: k3bcore.cpp 419869 2005-05-30 19:27:55Z trueg $
 * Copyright (C) 2003 Sebastian Trueg <trueg@k3b.org>
 *
 * This file is part of the K3b project.
 * Copyright (C) 1998-2004 Sebastian Trueg <trueg@k3b.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */

#include <config.h>

#include <k3bcore.h>

#include <k3bdevicemanager.h>
#include <k3bexternalbinmanager.h>
#include <k3bdefaultexternalprograms.h>
#include <k3bglobals.h>
#include <k3bversion.h>
#include <k3bjob.h>
#include <k3bthreadwidget.h>
#include <k3bglobalsettings.h>
#include <k3bpluginmanager.h>

#include <klocale.h>
#include <kconfig.h>
#include <kaboutdata.h>
#include <kstandarddirs.h>
#include <kapplication.h>

#include <qptrlist.h>

class K3bCore::Private
{
public:
  Private();
  Private(const Private& rhs);
  virtual ~Private();
  Private& operator= (const Private& rhs);

  K3bVersion version;
  KConfig* config;
  bool deleteConfig;
  K3bDevice::DeviceManager* deviceManager;
  K3bExternalBinManager* externalBinManager;
  K3bPluginManager* pluginManager;
  K3bGlobalSettings* globalSettings;

  QPtrList<K3bJob> runningJobs;
};

K3bCore::Private::Private()
  : version( LIBK3B_VERSION ),
    config(0L),
    deleteConfig(false),
    deviceManager(0L),
    externalBinManager(0L),
    pluginManager(0L),
    globalSettings(0L),
    runningJobs()
{
}

K3bCore::Private::Private(const K3bCore::Private& rhs)
  : version(rhs.version),
    config(rhs.config),
    deleteConfig(rhs.deleteConfig),
    deviceManager(rhs.deviceManager),
    externalBinManager(rhs.externalBinManager),
    pluginManager(rhs.pluginManager),
    globalSettings(rhs.globalSettings),
    runningJobs(rhs.runningJobs)
{
}

K3bCore::Private::~Private()
{
}

K3bCore::Private&
K3bCore::Private::operator= (const K3bCore::Private& rhs)
{
    if (this != &rhs)
    {
	version = rhs.version;
	config = rhs.config;
	deleteConfig = rhs.deleteConfig;
	deviceManager = rhs.deviceManager;
	externalBinManager = rhs.externalBinManager;
	pluginManager = rhs.pluginManager;
	globalSettings = rhs.globalSettings;
	runningJobs = rhs.runningJobs;
    }
    return *this;
}

K3bCore* K3bCore::s_k3bCore = 0L;
QMutex* K3bCore::m_mutex = new QMutex();

K3bCore::K3bCore( QObject* parent, const char* name )
  : QObject( parent, name ),
  d(0L)
{
  kdDebug() << "K3bCore::K3bCore(): Constructing a K3bCore Instance." << endl;
  kdDebug() << "K3bCore::K3bCore(): parent = 0x" << (void *) parent << " name = " << ((0L == name) ? "NULL" : name) << endl;
  d = new Private();
  kdDebug() << "K3bCore::K3bCore(): d = 0x" << (void *) d << endl;

  if( K3bCore::s_k3bCore )
  {
    qFatal("ONLY ONE INSTANCE OF K3BCORE ALLOWED!");
    kdDebug() << "K3bCore::K3bCore(): ONLY ONE INSTANCE OF K3BCORE IS ALLOWED!" << endl;
  }

  // ORIGINAL: WAS HERE
  K3bCore::s_k3bCore = this;

  d->globalSettings = new K3bGlobalSettings();
  kdDebug() << "K3bCore::K3bCore(): d->globalSettings = 0x" << (void *) d->globalSettings << endl;
  d->externalBinManager = new K3bExternalBinManager( this );
  kdDebug() << "K3bCore::K3bCore(): d->externalBinManager = 0x" << (void *) d->externalBinManager << endl;
  d->deviceManager = new K3bDevice::DeviceManager( this );
  kdDebug() << "K3bCore::K3bCore(): d->deviceManager = 0x" << (void *) d->deviceManager << endl;
  d->pluginManager = new K3bPluginManager( this );
  kdDebug() << "K3bCore::K3bCore(): d->pluginManager = 0x" << (void *) d->pluginManager << endl;

  // K3bCore::s_k3bCore = this;
  kdDebug() << "K3bCore::K3bCore(): K3bCore::s_k3bCore = 0x" << (void *) K3bCore::s_k3bCore << " this = 0x" << (void *) this << endl;

  K3b::addDefaultPrograms( d->externalBinManager );

  // create the thread widget instance in the GUI thread
  K3bThreadWidget::instance();
  kdDebug() << "K3bCore::K3bCore(): K3bCore Instance Constructed." << endl;
}


K3bCore::~K3bCore()
{
  kdDebug() << "K3bCore::~K3bCore(): Destructing a K3bCore Instance." << endl;
  delete d->globalSettings;
  delete d;
  d = 0L;
  kdDebug() << "K3bCore::~K3bCore(): K3bCore Instance Destructed." << endl;
}

K3bCore* K3bCore::k3bCore()
{
  kdDebug() << "K3bCore::k3bCore(): K3bCore::s_k3bCore = 0x" << (void *) K3bCore::s_k3bCore << endl;
  if (0L == K3bCore::s_k3bCore)
  {
    QMutexLocker mutexLock(K3bCore::m_mutex);
    if (0L == K3bCore::s_k3bCore)
        K3bCore::s_k3bCore = new K3bCore();
  }
  return K3bCore::s_k3bCore;
}


K3bDevice::DeviceManager* K3bCore::deviceManager() const
{
  return ((d) ? d->deviceManager : 0L);
}


K3bExternalBinManager* K3bCore::externalBinManager() const
{
  return ((d) ? d->externalBinManager : 0L);
}


K3bPluginManager* K3bCore::pluginManager() const
{
  kdDebug() << "K3bCore::pluginManager(): d = 0x" << (void *) d;
  if (d)
    kdDebug() << " d->pluginManager = 0x" << (void *) d->pluginManager << endl;

  return ((d) ? d->pluginManager : 0L);
}


K3bGlobalSettings* K3bCore::globalSettings() const
{
  return ((d) ? d->globalSettings : 0L);
}


const K3bVersion& K3bCore::version() const
{
  return d->version;
}


KConfig* K3bCore::config() const
{
  if( !d->config ) {
    kdDebug() << "(K3bCore) opening k3b config file." << endl;
    kdDebug() << "(K3bCore) while I am a " << className() << endl;
    d->deleteConfig = true;
    d->config = new KConfig( "k3brc" );
  }

  return d->config;
}


void K3bCore::init()
{
  // load the plugins before doing anything else
  // they might add external bins
  d->pluginManager->loadAll();

  d->externalBinManager->search();
  if( !d->deviceManager->scanBus() )
    kdDebug() << "No Devices found!" << endl;
}


void K3bCore::readSettings( KConfig* cnf )
{
  KConfig* c = cnf;
  if( !c )
    c = config();

  QString oldGrp = c->group();

  d->globalSettings->readSettings( c );
  d->deviceManager->readConfig( c );
  d->externalBinManager->readConfig( c );

  c->setGroup( oldGrp );
}


void K3bCore::saveSettings( KConfig* cnf )
{
  KConfig* c = cnf;
  if( !c )
    c = config();

  QString oldGrp = c->group();

  c->setGroup( "General Options" );
  c->writeEntry( "config version", version() );

  deviceManager()->saveConfig( c );
  externalBinManager()->saveConfig( c );
  d->globalSettings->saveSettings( c );

  c->setGroup( oldGrp );
}


void K3bCore::registerJob( K3bJob* job )
{
  d->runningJobs.append( job );
}


void K3bCore::unregisterJob( K3bJob* job )
{
  d->runningJobs.removeRef( job );
}


bool K3bCore::jobsRunning() const
{
  return !d->runningJobs.isEmpty();
}


const QPtrList<K3bJob>& K3bCore::runningJobs() const
{
  return d->runningJobs;
}

#include "k3bcore.moc"
