#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/fcntl.h>
#include <sys/param.h>

#include "k3bsolaris_device.h"

#ifdef __cplusplus
extern "C" {
#endif

static void printDiskIOInfo (const DiskIOInfo* dip);
static void printDiskSCSIInfo (const DiskSCSIInfo* dsp);
static void printKStat (const kstat_t* ksp);
static void printKIO (const kstat_io_t* kip);
static void printKIODelta (const DiskIOInfo* dip);
static void printKStatList (void);
static void fillDefaultValues (DiskSCSIInfo* dp);
static void copyDiskSCSIInfo (const kstat_t* ksp, const kstat_io_t* kip, DiskSCSIInfo* dp);
static void resetMemory (void);
static int dumpDeviceNode (di_node_t node, void* arg);
static int dumpMinorDeviceNode (di_node_t node, di_minor_t minor, void *arg);
static void walkDiskNodes (const char* driver, di_node_t node);
static void resetDDIInfo (void);
static int guessPropertyType(uchar_t *data, int len);
static int readDir (const char* diskPath, const char* svr4Path, char* partitionName);
static void setProperties (di_node_t node, DiskSCSIInfo* scsiInfo);
static void updateDDIInfo (di_node_t node, const char* driver);
static void printKStatList (void);
static int findInstance (int instance);
static void addInstance (int instance);
static void resetInstanceInfo (void);
static void setUTSName (void);
static void setDriverName (void);

static char driver[16];
static UTSName utsName;
static int hasDriver = 0;
static kstat_ctl_t* kc = 0L;
static kstat_t* ks = 0L;
static DiskSCSIInfo* scsiInfoPtr = 0L;
static DDIInfo* ddiInfoPtr = 0L;
static int* instancePtr = 0L;
static size_t numDDIInfo = 0;
static size_t numSCSIInfo = 0;
static size_t numInstances = 0;
static struct timeval tv;

void
setUTSName (void)
{
  if (uname (&(utsName.uName)) < 0)
  {
    perror("uname"), strerror(errno);
    utsName.ok = 0;
  }
  else
    utsName.ok = 1;
}

void
setDriverName (void)
{
  char* ptr;
  int version;

  if (!utsName.ok)
    setUTSName();

  hasDriver = 1;
}

void
printDiskIOInfo (const DiskIOInfo* dip)
{
  if (dip)
    {
      printKIO (&(dip->o));
      printKIO (&(dip->n));
    }
}

void
printDiskSCSIInfo (const DiskSCSIInfo* dsp)
{
  if (dsp)
  {
    (void) fprintf (stderr, "\tinstance = %ld\n", dsp->instance);
    (void) fprintf (stderr, "\tkid = %ld\n", dsp->kid);
    (void) fprintf (stderr, "\tlun = %ld\n", dsp->lun);
    (void) fprintf (stderr, "\ttarget = %ld\n", dsp->target);
    (void) fprintf (stderr, "\tdev = 0x%4.4x (%ld/%ld)\n", dsp->dev, major(dsp->dev), minor(dsp->dev));
    (void) fprintf (stderr, "\t\tmajor = %ld\n", major(dsp->dev));
    (void) fprintf (stderr, "\t\tminor = %ld\n", minor(dsp->dev));
    (void) fprintf (stderr, "\tkstatName = %s\n", dsp->kstatName);
    (void) fprintf (stderr, "\tkstatClass = %s\n", dsp->kstatClass);
    (void) fprintf (stderr, "\tvendorName = %s\n", dsp->vendorName);
    (void) fprintf (stderr, "\tproductId = %s\n", dsp->productId);
    (void) fprintf (stderr, "\trevision = %s\n", dsp->revision);
    (void) fprintf (stderr, "\tddiDeviceName = %s\n", dsp->ddiDeviceName);
    (void) fprintf (stderr, "\tspecType = %s\n", dsp->specType);
    (void) fprintf (stderr, "\tnodeType = %s\n", dsp->nodeType);
    (void) fprintf (stderr, "\tsvr4Device = %s\n", dsp->svr4Device);
    (void) fprintf (stderr, "\tscsiSlice = %s\n", dsp->scsiSlice);
    if (0 == strcmp (dsp->specType, "block"))
    {
      printDiskIOInfo (&(dsp->diskIOInfo));
      printKIODelta (&(dsp->diskIOInfo));
    }
  }
}

void
printKStat (const kstat_t* ksp)
{
  if (ksp)
    {
      (void) fprintf (stderr, "ks_module = %s\n", ksp->ks_module);
      (void) fprintf (stderr, "\tks_name = %s\n", ksp->ks_name);
      (void) fprintf (stderr, "\tks_instance = %ld\n", ksp->ks_instance);
      (void) fprintf (stderr, "\tks_class = %s\n", ksp->ks_class);
      (void) fprintf (stderr, "\tks_kid = %ld\n", ksp->ks_kid);
      (void) fprintf (stderr, "\tks_type = %ld\n", ksp->ks_type);
    }
}

void
printKIO (const kstat_io_t* kip)
{
    if (kip)
    {
        (void) fprintf (stderr, "\t\tnread = %llu\n", kip->nread); // u_longlong_t 
        (void) fprintf (stderr, "\t\tnwritten = %llu\n", kip->nwritten); // u_longlong_
        
        (void) fprintf (stderr, "\t\treads = %u\n", kip->reads); // uint_t
        (void) fprintf (stderr, "\t\twrites = %u\n", kip->writes); // uint_t
        
        (void) fprintf (stderr, "\t\twtime = %lld\n", kip->wtime ); // hrtime_t
        (void) fprintf (stderr, "\t\twlentime = %lld\n", kip->wlentime); // hrtime_t
        (void) fprintf (stderr, "\t\twlastupdate = %lld\n", kip->wlastupdate); // hrtime_t
        (void) fprintf (stderr, "\t\trtime = %lld\n", kip->rtime); // hrtime_t
        (void) fprintf (stderr, "\t\trlentime = %lld\n", kip->rlentime); // hrtime_t
        (void) fprintf (stderr, "\t\trlastupdate = %lld\n", kip->rlastupdate); // hrtime_t
        
        (void) fprintf (stderr, "\t\twcnt = %lld\n", kip->wcnt); // hrtime_t
        (void) fprintf (stderr, "\t\trcnt = %lld\n", kip->rcnt); // hrtime_t
    }
}

void
printKIODelta (const DiskIOInfo* dip)
{
  if (dip)
  {
    (void) fprintf (stderr, "\t\t\tbytes read: %llu\n", dip->n.nread - dip->o.nread);
    (void) fprintf (stderr, "\t\t\tbytes written: %llu\n", dip->n.nwritten - dip->o.nwritten);
    (void) fprintf (stderr, "\t\t\tread operations: %lu\n", dip->n.reads - dip->o.reads);
    (void) fprintf (stderr, "\t\t\twrite operations: %lu\n", dip->n.writes - dip->o.writes);
  }
}

void
fillDefaultValues (DiskSCSIInfo* dp)
{
  dp->instance = 0;
  dp->kid = 0;
  dp->lun = 0;
  dp->target = 0;
  dp->dev = 0x0;
  dp->major = 0;
  dp->minor = 0;
  (void) memset (dp->kstatName, '\0', sizeof (dp->kstatName));
  (void) memset (dp->kstatClass, '\0', sizeof (dp->kstatClass));
  (void) memset (dp->vendorName, '\0', sizeof (dp->vendorName));
  (void) memset (dp->productId, '\0', sizeof (dp->productId));
  (void) memset (dp->revision, '\0', sizeof (dp->revision));
  (void) memset (dp->ddiDeviceName, '\0', sizeof (dp->ddiDeviceName));
  (void) memset (dp->specType, '\0', sizeof (dp->specType));
  (void) memset (dp->nodeType, '\0', sizeof (dp->nodeType));
  (void) memset (dp->svr4Device, '\0', sizeof (dp->svr4Device));
  (void) memset (dp->scsiSlice, '\0', sizeof (dp->scsiSlice));
  (void) memset (&(dp->diskIOInfo), '\0', sizeof (struct DiskIOInfo));
}

void
copyDiskSCSIInfo (const kstat_t* ksp, const kstat_io_t* kip, DiskSCSIInfo* dp)
{
  if ((0L == ksp) || (0L == dp))
    return;
  
  dp->kid = ksp->ks_kid;
  (void) strcpy (dp->kstatName, ksp->ks_name);
  (void) strcpy (dp->kstatClass, ksp->ks_class);
  (void) memcpy (&(dp->diskIOInfo.o), &(dp->diskIOInfo.n), sizeof (kstat_io_t));
  (void) memcpy (&(dp->diskIOInfo.n), kip, sizeof (kstat_io_t));
}

void
resetMemory (void)
{
    if (scsiInfoPtr != 0)
    {
        free (scsiInfoPtr);
        scsiInfoPtr = 0L;
        numSCSIInfo = 0;
    }
}

int
updateDeviceAlias (void)
{
  char file[MAXPATHLEN];
  di_node_t rootNode;
  di_node_t node;
  di_minor_t minor = DI_MINOR_NIL;
  char tmpnode[MAXPATHLEN];
  char* phys_path;
  char* minor_name;
  int instance;

  resetDDIInfo();
  resetInstanceInfo();
  
  if ((rootNode = di_init("/", DINFOCPYALL)) == DI_NODE_NIL)
  {
      perror("di_init"), strerror(errno);
      return -1;
  }
  walkDiskNodes (driver, rootNode);
  node = di_drv_first_node(driver, rootNode);
  while (node != DI_NODE_NIL)
  {
      updateDDIInfo (node, driver);
      (void) di_walk_minor(node, NULL, 0, (void *) node, dumpMinorDeviceNode);
      node = di_drv_next_node(node);
  }
  di_fini(rootNode);
  return 0;
}

void
walkDiskNodes (const char* driver, di_node_t node)
{
    node = di_drv_first_node(driver, node);
    if (node == DI_NODE_NIL)
        return;

    updateDDIInfo (node, driver);
    node = di_drv_next_node(node);
    
    while (node != DI_NODE_NIL)
    {
        updateDDIInfo (node, driver);
        node = di_drv_next_node(node);
    }
}


int
updateKStatList (void)
{
    DiskSCSIInfo* diskInfoPtr = 0L;
    kstat_ctl_t* kc = 0L;
    kstat_t* ks = 0L;
    kstat_io_t kio;
    uint32_t i;
    
    if ((kc = kstat_open()) == (kstat_ctl_t *) NULL)
    {
        perror("kstat_open"), strerror(errno);
        return -1;
    }
    for (i = 0; i < numInstances; ++i)
    {
        if ((ks = kstat_lookup (kc, (char *) driver, instancePtr[i], NULL)) == (kstat_t *) NULL)
        {
            perror("kstat_lookup"), strerror(errno);
            continue;
        }
        
        while (kstat_chain_update(kc) != 0);
        
        do
        {
            if ((strcmp (ks->ks_module, driver) == 0) && (ks->ks_type == KSTAT_TYPE_IO))
            {
                if (kstat_read (kc, ks, &kio) == -1)
                {
                    perror("kstat_read"), strerror(errno);
                    kstat_close(kc);
                    return -1;
                }
                diskInfoPtr = findSCSISlice (ks->ks_name);
                if (0L == diskInfoPtr)
                {
                    while (kstat_chain_update(kc) != 0);
                    ks = ks->ks_next;
                    continue;
                }
                copyDiskSCSIInfo (ks, &kio, diskInfoPtr);
                while (kstat_chain_update(kc) != 0);
            }
            ks = ks->ks_next;
        } while (ks);
    }
    kstat_close(kc);
    return 0;
}

int
updateDeviceInfo (const char* drv)
{
    di_node_t rootNode;
    di_node_t node;
    di_prom_handle_t promHandle;
    int promProperties = 0;
    
    (void) memset(driver, '\0', (size_t) 16);
    (void) snprintf(driver, (size_t) 15, "%s", drv);
    if ((rootNode = di_init("/", DINFOCPYALL)) == DI_NODE_NIL)
    {
        perror("di_init"), strerror(errno);
        return -1;
    }
    if ((0 == getuid()) || (0 == geteuid()))
    {
        ++promProperties;
        if (promProperties && (promHandle = di_prom_init()) == DI_PROM_HANDLE_NIL)
        {
            perror("di_prom_init"), strerror(errno);
            di_fini(rootNode);
            return -1;
        }
    }
    node = di_drv_first_node(driver, rootNode);
    while (node != DI_NODE_NIL)
    {
        (void) di_walk_minor(node, NULL, 0, (void *) node, dumpMinorDeviceNode);
        if (promProperties)
          di_prom_fini(promHandle);
        node = di_drv_next_node(node);
    }
    di_fini(rootNode);
    return 0;
}

void
resetInstanceInfo (void)
{
    if (instancePtr)
    {
        free (instancePtr);
        instancePtr = 0L;
    }
}


void
resetDDIInfo (void)
{
    if (ddiInfoPtr)
    {
        free (ddiInfoPtr);
        ddiInfoPtr = 0L;
    }
}

void
updateDDIInfo (di_node_t node, const char* driver)
{
  int instance;
  char* physPath;

  if (di_minor_next(node, DI_MINOR_NIL) == DI_MINOR_NIL)
    return;

  instance = di_instance(node);
  physPath = di_devfs_path(node);
  
  ddiInfoPtr = (DDIInfo *) realloc (ddiInfoPtr, (numDDIInfo * sizeof(DDIInfo) + sizeof(DDIInfo)));
  if (0L == ddiInfoPtr)
    return;
  
  ddiInfoPtr[numDDIInfo].instance = instance;
  (void) strcpy (ddiInfoPtr[numDDIInfo].physicalPath, physPath);
  (void) strcpy (ddiInfoPtr[numDDIInfo].driver, driver);
  ++numDDIInfo;
}


int
dumpDeviceNode (di_node_t node, void* arg)
{
    char* name;
    di_prom_handle_t ph = (di_prom_handle_t)arg;
    
    name = di_devfs_path(node);
    di_devfs_path_free(name);

    if (di_minor_next(node, DI_MINOR_NIL) == DI_MINOR_NIL)
        return DI_WALK_TERMINATE;
    
    (void) di_walk_minor(node, NULL, 0, (void *) node, dumpMinorDeviceNode);
    return DI_WALK_CONTINUE;
}

int
dumpMinorDeviceNode (di_node_t node, di_minor_t minor, void *arg)
{
    char* name;
    char* specType;
    char* nodeType;
    char* physPath;
    char* minorName;
    dev_t dev;
    DiskSCSIInfo* scsiInfo;
    int instance;
    char svr4[10];
    char tmpNode[MAXPATHLEN];
    char file[MAXPATHLEN];
    char partition[MAXPATHLEN];
    
    name = di_devfs_path(node);
    di_devfs_path_free(name);
    
    if ((caddr_t) node != (caddr_t)arg)
        return DI_WALK_TERMINATE;
    
    instance = di_instance(node);
    physPath = di_devfs_path(node);
    minorName = di_minor_name(minor);
    (void) memset (partition, '\0', MAXPATHLEN);
    (void) memset (svr4, '\0', sizeof(svr4));
    
    (void) sprintf (tmpNode, "%s%d", driver, instance);
    (void) strcpy (file, "/devices");
    (void) strcat (file, physPath);
    (void) strcat (file, ":");
    (void) strcat (file, minorName);
    
    specType = (di_minor_spectype(minor) == S_IFCHR) ? "char" : "block";
    nodeType = di_minor_nodetype(minor);
    
    if (di_minor_spectype(minor) == S_IFBLK)
        (void) strcpy (svr4, "/dev/dsk");
    else if (di_minor_spectype(minor) == S_IFCHR)
        (void) strcpy (svr4, "/dev/rdsk");
    else
    {
        return DI_WALK_TERMINATE;
    }
    if (0 == readDir(file, svr4, partition))
    {
        scsiInfo = findDiskSCSIInfo(file);
        if (0L == scsiInfo)
        {
            scsiInfoPtr = (DiskSCSIInfo *) realloc (scsiInfoPtr,
                                                    (numSCSIInfo * sizeof(struct DiskSCSIInfo)) + sizeof(struct DiskSCSIInfo));
            if (0L == scsiInfoPtr)
                return DI_WALK_TERMINATE;
            
            fillDefaultValues (&(scsiInfoPtr[numSCSIInfo]));
            scsiInfo = &(scsiInfoPtr[numSCSIInfo]);
            ++numSCSIInfo;
        }
        scsiInfo->instance = instance;
        addInstance(scsiInfo->instance);
        scsiInfo->dev = di_minor_devt(minor);
        if (scsiInfo->dev != DDI_DEV_T_NONE)
        {
            scsiInfo->major = major(scsiInfo->dev);
            scsiInfo->minor = minor(scsiInfo->dev);
        }
        (void) strcpy (scsiInfo->ddiDeviceName, file);
        (void) strcpy (scsiInfo->svr4Device, partition);
        (void) strcpy (scsiInfo->specType, specType);
        (void) strcpy (scsiInfo->nodeType, ((nodeType) ? nodeType : ""));
        (void) sprintf (scsiInfo->scsiSlice, "%s%d,%s", driver, instance, di_minor_name(minor));
        setProperties (node, scsiInfo);
    }
    return DI_WALK_CONTINUE;
}

void
setProperties (di_node_t node, DiskSCSIInfo* scsiInfo)
{
    int rval, i, type;
    di_prop_t prop, next;
    char buf[256];
    int* intp;
    char* strp;
    char* ptr;
    uchar_t* bytep;
    
    if ((next = di_prop_next(node, DI_PROP_NIL)) == DI_PROP_NIL)
    {
        return;
    }
    while (next != DI_PROP_NIL)
    {
        prop = next;
        next = di_prop_next(node, prop);
        type = di_prop_type(prop);
        if (type == DI_PROP_TYPE_UNKNOWN)
        {
            rval = di_prop_bytes(prop, &bytep);
            type = guessPropertyType(bytep, rval);
        }
        if (type == DI_PROP_TYPE_INT)
        {
            if ((strcmp (di_prop_name(prop), "lun") == 0) || (strcmp (di_prop_name(prop), "target") == 0))
            {
                rval = di_prop_ints(prop, &intp);
                if (rval == -1)
                {
                    perror("di_prop_ints"), strerror(errno);
                    break;
                }
                (void) memset (buf, '\0', sizeof(buf));
                for (i = 0; i < rval; ++i)
                {
                    ptr = &buf[i];
                    (void) sprintf (ptr, "%d", intp[i]);
                }
                if (strcmp(di_prop_name(prop), "lun") == 0)
                    scsiInfo->lun = atoi(buf);
                else
                    scsiInfo->target = atoi(buf); 
            }
        }
        else if (type == DI_PROP_TYPE_STRING)
        {
            if ((strcmp (di_prop_name(prop), "inquiry-vendor-id") == 0) ||
                (strcmp (di_prop_name(prop), "inquiry-product-id") == 0) ||
                (strcmp (di_prop_name(prop), "inquiry-revision-id") == 0))
            {
                rval = di_prop_strings(prop, &strp);
                if (rval == -1)
                {
                    perror("di_prop_strings"), strerror(errno);
                    break;
                }
                if (strcmp (di_prop_name(prop), "inquiry-vendor-id") == 0)
                    (void) sprintf (scsiInfo->vendorName, "%s", strp);
                else if (strcmp (di_prop_name(prop), "inquiry-product-id") == 0)
                    (void) sprintf (scsiInfo->productId, "%s", strp);
                else
                    (void) sprintf (scsiInfo->revision, "%s", strp);
                strp += strlen(strp) + 1;
            }
        }
    }
}

int
guessPropertyType(uchar_t *data, int len)
{
    int i, c, slen, guess;
    
    if (len < 0)
        return (-1);
    else if (len == 0)
        return (DI_PROP_TYPE_BOOLEAN);
    
    slen = 0;
    guess = DI_PROP_TYPE_STRING;
    
    for (i = 0; i < len; i++)
    {
        c = (int)data[i];
        switch (c)
        {
            case 0: /* null character */
                if (slen == 0)
                    guess = DI_PROP_TYPE_BYTE;
                else
                    slen = 0;
                break;
            default:
                if (!isprint(c))
                    guess = DI_PROP_TYPE_BYTE;
                else
                    slen++;
        }
        if (guess != DI_PROP_TYPE_STRING)
            break;
    }
    if ((guess == DI_PROP_TYPE_BYTE) && (len % sizeof(int) == 0))
        guess = DI_PROP_TYPE_INT;
    
    return (guess);
}

int
readDir (const char* diskPath, const char* svr4Path, char* partitionName)
{
    DIR* dirp;
    struct dirent *dp;
    struct stat stbuf;
    char fileName[MAXPATHLEN];
    char tempName[MAXPATHLEN];
    char dirDname[MAXPATHLEN];
    
    if ((dirp = opendir(svr4Path)) != NULL)
    {
        while ((dp = readdir(dirp)) != NULL)
        {
            (void) sprintf(tempName, "../..%s", diskPath);
            (void) sprintf(dirDname, "%s/%s", svr4Path, dp->d_name);
            stat (dirDname, &stbuf);

            if (0 == strcmp (svr4Path, "/dev/dsk"))
            {
                if (S_ISBLK(stbuf.st_mode))
                {
                    readlink(dirDname, fileName, sizeof(fileName));
                    if (strcmp(fileName, tempName) == 0)
                    {
                        (void) strcpy (partitionName, dirDname);
                        closedir(dirp);
                        return 0;
                    }
                }
            }
            else if (0 == strcmp (svr4Path, "/dev/rdsk"))
            {
                if (S_ISCHR(stbuf.st_mode))
                {
                    readlink(dirDname, fileName, sizeof(fileName));
                    if (strcmp(fileName, tempName) == 0)
                    {
                        (void) strcpy (partitionName, dirDname);
                        closedir(dirp);
                        return 0;
                    }
                }
            }
            (void) memset (fileName,'\0', MAXPATHLEN);
            (void) memset (tempName,'\0', MAXPATHLEN);
            (void) memset (dirDname,'\0', MAXPATHLEN);
        }
        closedir(dirp);
    }
    return(-1);
}

DiskSCSIInfo*
findDiskSCSIInfo (const char* ddiDeviceName)
{
    uint32_t i;
    
    if (0L == ddiDeviceName)
        return 0L;
    
    for (i = 0; i < numSCSIInfo; ++i)
    {
        if (0 == strcmp (scsiInfoPtr[i].ddiDeviceName, ddiDeviceName))
            return (&(scsiInfoPtr[i]));
    }
    return 0L;
}

DDIInfo*
findDDIInfo (const char* ddiDeviceName)
{
    uint32_t i;

    if (0L == ddiDeviceName)
        return 0L;
    
    for (i = 0; i < numDDIInfo; ++i)
    {
        if (0 == strcmp (ddiInfoPtr[i].physicalPath, ddiDeviceName))
            return (&(ddiInfoPtr[i]));
    }
    return 0L;
}

int
findInstance (int instance)
{
    uint32_t i;
    
    for (i = 0; i < numInstances; ++i)
        if (instancePtr[i] == instance)
            return instancePtr[i];

    return -1;
}

DiskSCSIInfo*
findSCSISlice (const char* sliceName)
{
    uint32_t i;
    
    if (0L == sliceName)
        return 0L;
    
    for (i = 0; i < numSCSIInfo; ++i)
        if (strcmp (scsiInfoPtr[i].scsiSlice, sliceName) == 0)
            return (&(scsiInfoPtr[i]));

    return 0L;
}

DiskSCSIInfo*
findSVR4Device (const char* deviceName)
{
  uint32_t i;

  if (0L == deviceName)
    return 0L;

  for (i = 0; i < numSCSIInfo; ++i)
    if (strcmp(scsiInfoPtr[i].svr4Device, deviceName) == 0)
      return (&(scsiInfoPtr[i]));

  return 0L;
}


void
addInstance (int instance)
{
    if (findInstance (instance) < 0)
    {
        instancePtr = (int *) realloc (instancePtr, sizeof(int) * numInstances + sizeof(int));
        if (instancePtr)
        {
            instancePtr[numInstances] = instance;
            ++numInstances;
        }
    }
}

void
printKStatList(void)
{
    size_t i;
    if (numSCSIInfo)
    {
        for (i = 0; i < numSCSIInfo; ++i)
            printDiskSCSIInfo (&(scsiInfoPtr[i]));
    }
}

#ifdef __cplusplus
} // extern "C"
#endif

// vim: ts=2 sw=2 et

